/************************************************************************/
/*  msfix -	fix (remove) corrupt records in MiniSEED file.		*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: msfix.c,v 1.5 2004/01/25 06:36:49 doug Exp $ ";
#endif

#include <stdio.h>
#include "version.h"

char *syntax[] = {
"%s version " VERSION,
"%s  [-h] [-n] [-v] [-m] [-D j|m] [ infile [ outfile ]]",
"    where:",
"	-h	Help - prints syntax message.",
"	-n	No output.",
"	-v	Verbose output.",
"	-m	Match any SNCL, not just SNCL of the volhdr or first record.",
"		Keep good records from any data channel.",
"	-D j|m	Date format - julian (day of year) or month format.",
"	infile	Optional input file containing MiniSEED data.",
"		If no input file is specified, data is read from stdin.",
"	outfile	Optional output file for fixed MiniSEED data.",
"Notes:",
"1.  If no input file is specified, read from stdin and write to stdout.",
"2.  If only input file is specified, replace contents of input file with",
"    the fixed MiniSEED output data.",
"3.  If both input and output files are specified, write fixed output data",
"    to output file.",
"4.  If outfile = '-', write output to stdout.",
"5.  If -n is specified, do not write any output.",
"Examples:",
"1.  %s PKD.BK.BHZ..D.2004.024.0000",
"	Fix and rewrite the specified file.",
"2.  %s PKD.BK.BHZ..D.2004.024.0000 newfile",
"	Fix the data from the input file and write output to newfile.",
"3.  %s PKD.BK.BHZ..D.2004.024.0000 - ",
"	Fix the data from the input file and write output to stdout.",
"4.  %s PKD.BK.BHZ..D.2004.024.0000 - | sdrsplit -C ",
"	Fix the data from the input file and write output to stdout pipe",
"	which is input to sdrsplit.",
NULL };

#include <string.h>
#include <errno.h>
#ifdef	sun
#include <libgen.h>
#endif
#ifdef	linux
#include <libgen.h>
#endif

#include <qlib2.h>

#define	MSMAX	8192

char *cmdname;
FILE *info;

static FILE *input, *output;
static char *infile, *outfile;
static int replace_file = 0;
static int no_output = 0;

/************************************************************************/
/*  main routine							*/
/************************************************************************/
main (int argc, char **argv)
{
    DATA_HDR *hdr = NULL;
    char buf[MSMAX];
    char station[DH_STATION_LEN+1];
    char channel[DH_CHANNEL_LEN+1];
    char network[DH_NETWORK_LEN+1];
    char location[DH_LOCATION_LEN+1];
    char tempstr[2048];
    char *outdir;
    
    int *data;
    int i;
    int status;
    int blksize = 0;
    int inrecnum = 0;
    int outrecnum = 0;
    int verbose = 0;
    int offset = 0;
    int nbytes = 0;
    int date_fmt = JULIAN_FMT;
    int match_sncl = 1;
    
    int m;
    double rate;

    /* Variables needed for getopt. */
    extern char	*optarg;
    extern int	optind, opterr;
    int		c;

    cmdname = tail(*argv);
    info = stdout;
    /*	Parse command line options.					*/
    while ( (c = getopt(argc,argv,"hnvmD:")) != EOF)
	switch (c) {
	case '?':
	case 'h':   print_syntax (cmdname, syntax, info); exit(0); break;
	case 'n':   no_output = 1; break;
	case 'v':   verbose = 1; break;
	case 'm':   match_sncl = 0; break;
	case 'D':
	    /* Date display format.   */
	    if (strcmp (optarg, "j") == 0) date_fmt = JULIAN_FMT;
	    else if (strcmp (optarg, "j,") == 0) date_fmt = JULIAN_FMT_1;
	    else if (strcmp (optarg, "m") == 0) date_fmt = MONTH_FMT;
	    else if (strcmp (optarg, "m,") == 0) date_fmt = MONTH_FMT_1;
	    else {
		fprintf (info, "Error: invalid format for date display: %s\n", optarg);
		exit(1);
	    }
	    break;
	default:    fprintf (info, "Error: Invalid option: -%c\n", c); exit(1);
	}

    /*	Skip over all options and their arguments.			*/
    argv = &(argv[optind]);
    argc -= optind;

    /* The remaining arguments are [ input [and output] ] files.	*/

    infile = "<stdin>";
    input = stdin;
    outfile = "<stdout>";
    output = stdout;
    switch (argc) {
      case 0:	
	if (no_output) {
	    output = NULL;
	}
	else {
	    info = stderr;
	}
	break;
      case 1:
	infile = argv[0];
	if ((input = fopen (infile, "r")) == NULL) {
	    fprintf (info, "Error: unable to open input file %s\n", infile);
	    exit(1);
	}
	if (no_output) {
	    outfile = NULL;
	}
	else {
	    replace_file = 1;
	    strcpy (tempstr,infile);
	    outdir = dirname(tempstr);
	    outfile = tempnam (outdir,cmdname);
	    if ((output = fopen (outfile, "w")) == NULL) {
		fprintf (info, "Error: unable to open output file %s\n", outfile);
		exit(1);
	    }
	}
	break;
      case 2: 
	if (no_output) {
	    fprintf (info, "Error: conflicting options -n and output filename\n");
	    exit(1);
	}
	infile = argv[0];
	if ((input = fopen (infile, "r")) == NULL) {
	    fprintf (info, "Error: unable to open input file %s\n", infile);
	    exit(1);
	}
	outfile = argv[1];
	if (strcmp(outfile, "-") == 0) {
	    output = stdout;
	    info = stderr;
	}
	else {
	    if ((output = fopen (outfile, "w")) == NULL) {
		fprintf (info, "Error: unable to open output file %s\n", outfile);
		exit(1);
	    }
	}
	break;
      default:
	print_syntax (cmdname, syntax, info); exit(1); break;
    }

    /* Read the MiniSEED data.						*/
    station[0] = '\0';
    nbytes = fread (buf, 1, MSMAX, input);
    if (ferror(input)) {
	fprintf (info, "Error: reading input file %s\n", infile);
	cleanout(1,1);
	exit(1);
    }

    /* Loop over all records in the file. */
    while (nbytes > 0) {
	hdr = decode_hdr_sdr ((SDR_HDR *)(buf+offset), nbytes-offset);
	if (hdr == NULL && blksize == 0) {
	    fprintf (info, "Error: unable to decode record from input file %s\n", infile);
	    cleanout(1,1);
	    exit(1);
	}
	if (blksize == 0) {
	    /* Determine blksize from first record in file. */
	    blksize = hdr->blksize;
	    if (blksize == 0) {
		fprintf (info, "Error: unable to determine blksize from input file %s\n", infile);
		cleanout(1,1);
		exit(1);
	    }
	    if (verbose) {
		fprintf (info, "File: %s blksize = %d\n", infile, blksize);
		fflush (info);
	    }
	}
	if (hdr == NULL) {
	    fprintf (info, "Error: Record %d - Invalid record - SKIP\n", inrecnum);
	    offset += blksize;
	    ++inrecnum;
	}
	else if (hdr->blksize != blksize) {
	    fprintf (info, "Error: Record %d - Input record size %d not expected record size %d - SKIP\n",
		     inrecnum, hdr->blksize, blksize);
	    offset += blksize;
	    ++inrecnum;
	}
	else if (hdr->record_type == VOL_HDR_IND && outrecnum == 0) {
	    /* Copy volume hdr to output file. */
	    if (verbose) {
		fprintf (info, "record %d seq=%d volhdr\n", inrecnum, hdr->seq_no);
		fflush (info);
	    }
	    if (! no_output) {
		m = fwrite (buf+offset, blksize, 1, output);
		if (m != 1) {
		    fprintf (info, "Error: writing output file %s\n", outfile);
		    cleanout(1,1);
		    exit(1);
		}
	    }
	    offset += blksize;
	    ++inrecnum;
	    ++outrecnum;
	}
	else if (hdr->record_type == DATA_HDR_IND) {
	    /* We have a record or volume header. */
	    /* Get initial data stream info if we can. */
	    if (station[0] == '\0') {
		strcpy (station, hdr->station_id);
		strcpy (channel, hdr->channel_id);
		strcpy (network, hdr->network_id);
		strcpy (location, hdr->location_id);
	    }
	    if (match_sncl && 
		(strcmp(station,hdr->station_id) ||
		 strcmp(channel,hdr->channel_id) ||
		 strcmp(network,hdr->network_id) ||
		 strcmp(location,hdr->location_id))) {
		fprintf (info, "Error: Record %d - Expected %s.%s.%s.%s found %s.%s.%s.%s - SKIP\n",
			 inrecnum, station, network, channel, location,
			 hdr->station_id, hdr->network_id, hdr->channel_id, hdr->location_id);
		++inrecnum;
		offset += blksize;
	    }
	    else {
		/* Copy record to output file. */
		if (verbose) {
		    fprintf (info, "record %d seq=%d %s.%s.%s.%s %s",
			     inrecnum, hdr->seq_no,
			     hdr->station_id, hdr->network_id, hdr->channel_id, hdr->location_id,
			     time_to_str(hdr->begtime, date_fmt));
		    if (hdr->num_samples > 0 && 
			(rate=sps_rate(hdr->sample_rate,hdr->sample_rate_mult)) != 0) {
			INT_TIME endtime;
			int seconds, usecs;
			time_interval2 (1, hdr->sample_rate, hdr->sample_rate_mult, &seconds, &usecs);
			endtime = add_time (hdr->endtime,seconds,usecs);
			fprintf (info, " to %s\n", time_to_str(endtime, date_fmt));
		    }
		    else {
			fprintf (info, "\n");
		    }
		    fflush (info);
		}
		if (! no_output) {
		    m = fwrite (buf+offset, blksize, 1, output);
		    if (m != 1) {
			fprintf (info, "Error: writing output file %s\n", outfile);
			cleanout(1,1);
			exit(1);
		    }
		}
		offset += blksize;
		++inrecnum;
		++outrecnum;
	    }
	}
	else {
	    fprintf (info, "Error: Record %d - Unexpected record type: %c - SKIP\n",
		     inrecnum, hdr->record_type);
	    offset += blksize;
	    ++inrecnum;
	}

	if (hdr) {
	    free_data_hdr (hdr);
	    hdr = NULL;
	}

	/* Check to see if we need to read more data. */
	if (offset > nbytes) {
	    fprintf (info, "Error: - internal accounting error : offset %d > nbytes %d\n",
		     offset, nbytes);
	    cleanout(1,1);
	    exit(1);
	}
	if (offset < nbytes && offset + blksize > nbytes) {
	    fprintf (info, "Error: - internal accounting error : offset %d + blksize %d > nbytes %d\n",
		     offset, blksize, nbytes);
	    cleanout(1,1);
	    exit(1);
	}
	if (offset == nbytes) {
	    nbytes = 0;
	    offset = 0;
	}
	if (nbytes == 0) {
	    nbytes = fread (buf, 1, MSMAX, input);
	    if (ferror(input)) {
		fprintf (info, "Error: reading input file %s\n", infile);
		cleanout(1,1);
		exit(1);
	    }
	}
    }

    if ((status = fclose (input)) != 0) {
	fprintf (info, "Error: closing input file %s\n", infile);
	cleanout(1,1);
	exit(1);
    }
    if ((! no_output) && (status = fclose (output)) != 0) {
	fprintf (info, "Error: closing output file %s\n", outfile);
	cleanout(1,1);
	exit(1);
    }

    if (replace_file) {
	if (inrecnum != outrecnum) {
	    status = rename (outfile, infile);
	    if (status != 0) {
		fprintf (info, "Error: unable to rename output file %s to %s - %s\n",
			 outfile, infile, strerror(errno));
		cleanout(0,1);
		exit(errno);
	    }
	}
	else {
	    cleanout(0,1);
	}
    }
		
    fprintf (info, "%s.%s.%s.%s - %d input record, %d output records\n",
	     station, network, channel, location, inrecnum, outrecnum);
    return (0);
}

int cleanout (int close_flag, int unlink_flag)
{
    int status = 0;
    if (close_flag && (! no_output) && (status = fclose (output)) != 0) {
	fprintf (info, "Error: closing output file %s\n", outfile);
    }
    if (unlink_flag && replace_file && (status = unlink (outfile)) != 0) {
	fprintf (info, "Error: unlinking file %s - %s\n",
		 outfile, strerror(errno));
    }
    return (0);
}
